import io
import json
import logging
from typing import Any, AsyncGenerator, Dict, Optional, Tuple, Type

import aiohttp
from aiohttp.client_exceptions import ClientResponseError
from graphql import DocumentNode, ExecutionResult, print_ast
from multidict import CIMultiDictProxy

from gql.transport import AsyncTransport
from gql.utils import extract_files
from gql.transport.exceptions import (
    TransportClosed,
    TransportProtocolError,
    TransportServerError,
)

from ..auth import AbstractAsyncSession

log = logging.getLogger(__name__)

class AuthenticatedAIOHTTPTransport(AsyncTransport):
    """:ref:`Async Transport <async_transports>` to execute GraphQL queries
    on remote servers with an authenticated HTTP connection.

    This transport use the aiohttp library with asyncio.
    """

    file_classes: Tuple[Type[Any], ...] = (
        io.IOBase,
        aiohttp.StreamReader,
        AsyncGenerator,
    )

    def __init__(
        self,
        url: str,
        session: AbstractAsyncSession = None
    ) -> None:
        """Initialize the transport with the given aiohttp parameters.

        :param url: The GraphQL server URL. Example: 'https://server.com:PORT/path'.
        :param session: The authenticated session to utilize

        """
        self.url: str = url
        self.session: AbstractAsyncSession = session
        self.response_headers: Optional[CIMultiDictProxy[str]]

    async def connect(self) -> None:
        """Coroutine which will create an aiohttp ClientSession() as self.session.

        Don't call this coroutine directly on the transport, instead use
        :code:`async with` on the client and this coroutine will be executed
        to create the session.

        Should be cleaned with a call to the close coroutine.
        """
        # session management not performed by this transport

    async def close(self) -> None:
        """Coroutine which will close the aiohttp session.

        Don't call this coroutine directly on the transport, instead use
        :code:`async with` on the client and this coroutine will be executed
        when you exit the async context manager.
        """
        # session management not performed by this transport

    async def execute(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        extra_args: Dict[str, Any] = None,
        upload_files: bool = False,
    ) -> ExecutionResult:
        """Execute the provided document AST against the configured remote server
        using the current session.
        This uses the aiohttp library to perform a HTTP POST request asynchronously
        to the remote server.

        Don't call this coroutine directly on the transport, instead use
        :code:`execute` on a client or a session.

        :param document: the parsed GraphQL request
        :param variable_values: An optional Dict of variable values
        :param operation_name: An optional Operation name for the request
        :param extra_args: additional arguments to send to the aiohttp post method
        :param upload_files: Set to True if you want to put files in the variable values
        :returns: an ExecutionResult object.
        """

        query_str = print_ast(document)

        payload: Dict[str, Any] = {
            "query": query_str,
        }

        if operation_name:
            payload["operationName"] = operation_name

        if upload_files:

            # If the upload_files flag is set, then we need variable_values
            assert variable_values is not None

            # If we upload files, we will extract the files present in the
            # variable_values dict and replace them by null values
            nulled_variable_values, files = extract_files(
                variables=variable_values,
                file_classes=self.file_classes,
            )

            # Save the nulled variable values in the payload
            payload["variables"] = nulled_variable_values

            # Prepare aiohttp to send multipart-encoded data
            data = aiohttp.FormData()

            # Generate the file map
            # path is nested in a list because the spec allows multiple pointers
            # to the same file. But we don't support that.
            # Will generate something like {"0": ["variables.file"]}
            file_map = {str(i): [path] for i, path in enumerate(files)}

            # Enumerate the file streams
            # Will generate something like {'0': <_io.BufferedReader ...>}
            file_streams = {str(i): files[path] for i, path in enumerate(files)}

            # Add the payload to the operations field
            operations_str = json.dumps(payload)
            log.debug("operations %s", operations_str)
            data.add_field(
                "operations", operations_str, content_type="application/json"
            )

            # Add the file map field
            file_map_str = json.dumps(file_map)
            log.debug("file_map %s", file_map_str)
            data.add_field("map", file_map_str, content_type="application/json")

            # Add the extracted files as remaining fields
            for k, v in file_streams.items():
                data.add_field(k, v, filename=getattr(v, "name", k))

            post_args: Dict[str, Any] = {"data": data}

        else:
            if variable_values:
                payload["variables"] = variable_values

            if log.isEnabledFor(logging.INFO):
                log.debug(">>> %s", json.dumps(payload))

            post_args = {"json": payload}

        # Pass post_args to aiohttp post method
        if extra_args:
            post_args.update(extra_args)

        if self.session is None:
            raise TransportClosed("Transport is not connected")

        async with await self.session.post(self.url, **post_args) as resp:

            async def raise_response_error(resp: aiohttp.ClientResponse, reason: str):
                # We raise a TransportServerError if the status code is 400 or higher
                # We raise a TransportProtocolError in the other cases

                try:
                    # Raise a ClientResponseError if response status is 400 or higher
                    resp.raise_for_status()
                except ClientResponseError as e:
                    raise TransportServerError(str(e), e.status) from e

                result_text = await resp.text()
                raise TransportProtocolError(
                    f"Server did not return a GraphQL result: "
                    f"{reason}: "
                    f"{result_text}"
                )

            try:
                result = await resp.json(content_type=None)

                if log.isEnabledFor(logging.INFO):
                    result_text = await resp.text()
                    log.debug("<<< %s", result_text)

            except Exception:
                await raise_response_error(resp, "Not a JSON answer")

            if "errors" not in result and "data" not in result:
                await raise_response_error(resp, 'No "data" or "errors" keys in answer')

            # Saving latest response headers in the transport
            self.response_headers = resp.headers

            return ExecutionResult(
                errors=result.get("errors"),
                data=result.get("data"),
                extensions=result.get("extensions"),
            )

    def subscribe(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
    ) -> AsyncGenerator[ExecutionResult, None]:
        """Subscribe is not supported on HTTP.

        :meta private:
        """
        raise NotImplementedError(" The HTTP transport does not support subscriptions")


""" Implements support for the MilaCares GraphQL API """

import asyncio
import logging
from pathlib import Path
from typing import Any, Dict, Optional

from graphql import DocumentNode, ExecutionResult
from gql import Client
from gql.dsl import *
from gql.transport.exceptions import (
    TransportQueryError,
    TransportProtocolError,
    TransportServerError,
)

from milasdk import (
    API_DEFAULT_429_WAIT,
    API_BASE_URL,
    AbstractAsyncSession,
    MilaError
)
from milasdk.exceptions import OAuthError

from .gql import *

_LOGGER = logging.getLogger(__name__)

class MilaApi:
    """ A class that provides API calling functionality to the Mila API """

    def __init__(self, session: AbstractAsyncSession) -> None:
        self._session = session
        self._transport = AuthenticatedAIOHTTPTransport(API_BASE_URL, session)
        self._client = self._setup_client()

    def _setup_client(self) -> Client:
        with open(Path(__file__).parent / './gql/mila_schema.gql') as f:
            schema_str = f.read()

        c = Client(
            schema=schema_str,
            transport=self._transport, 
            serialize_variables=True, 
            parse_results=True
        )

        register_enums(c)
        register_scalers(c)

        return c

    async def _execute(self, document: DocumentNode, variable_values: Optional[Dict[str,Any]] = None) -> ExecutionResult:
        """ Main function to call the Mila API over HTTPS """
        retry = 3
        authError = 0
        response: ExecutionResult = None
        async with self._client as session:
            while retry:
                retry -= 1
                try:
                    return await session.execute(document, variable_values)
                except TransportServerError as ex:
                    if ex.code == 429:    # Too Many Requests
                        wait_time = response.headers.get('Retry-After', API_DEFAULT_429_WAIT)
                        _LOGGER.debug('HTTP Error 429 - Too Many Requests. Sleeping for %s seconds and will retry', wait_time)
                        await asyncio.sleep(int(wait_time)+1)
                    elif ex.code == 401 or ex.code >= 500: # Unauthorized or service error
                        if ex.code == 401:
                            authError += 1
                        # This is probably caused by an expired token so the next retry will get a new one automatically
                        _LOGGER.debug("API call returned error code=%d - %d retries left", ex.code, retry)
                    else:
                        raise MilaError("Transport server error occurred") from ex
                except TransportProtocolError as ex:
                    raise MilaError("Transport protocol error occurred") from ex
                except TransportQueryError as ex:
                    raise MilaError("Transport reported query error") from ex
                except GraphQLError as ex:
                    raise MilaError("Invalid query") from ex
                except OAuthError:
                    raise
                except Exception as ex:                
                    _LOGGER.debug("Unknown error occurred", exc_info=ex)
                    if not retry:
                        raise MilaError("API call failed") from ex

        # if we have multiple auth errors, assume we have an issue so we can
        # get a new token
        if authError > 1:
            raise OAuthError("Multiple auth errors, assuming auth token invalid")

        # all retries were exhausted without a valid response
        raise MilaError("Failed to get a valid response from Mila API service")

    async def get_account(self) -> Dict[str, Any]:
        """ Returns the account information (account, home, etc) """
        ds = DSLSchema(self._client.schema)
        query = dsl_gql(
            DSLQuery(
                ds.Query.owner.select(
                    ds.Owner.profile.select(
                        ds.Profile.email,
                        ds.Profile.firstName,
                        ds.Profile.lastName
                    )
                )
            )
        )
        result = await self._execute(query)

        return result["owner"]["profile"]

    async def get_appliances(self) -> Dict[str, Any]:
        """ Returns the information for all appliances """
        ds = DSLSchema(self._client.schema)
        query = dsl_gql(
            DSLQuery(
                ds.Query.owner.select(
                    ds.Owner.appliances.select(
                        *appliance_fields_fragment(ds)
                    )
                )
            )
        )
        result = await self._execute(query)
        return result["owner"]["appliances"]

    async def get_appliance(self, device_id: str) -> Dict[str, Any]:
        """ Returns information for the selected appliance """
        ds = DSLSchema(self._client.schema)
        query = dsl_gql(
            DSLQuery(
                ds.Query.owner.select(
                    ds.Owner.appliance(applianceId=device_id).select(
                        *appliance_fields_fragment(ds)
                    )
                )
            )
        )
        result = await self._execute(query)
        return result["owner"]["appliance"]

    async def get_appliance_sensor(self, device_id: str, sensor_kind: ApplianceSensorKind) -> Dict[str, Any]:
        """ Returns information for the selected appliance """
        ds = DSLSchema(self._client.schema)
        query = dsl_gql(
            DSLQuery(
                ds.Query.owner.select(
                    ds.Owner.appliance(applianceId=device_id).select(
                        ds.Appliance.sensor(kind=sensor_kind).select(
                            *appliance_sensor_fields_fragment(ds)
                        )
                    )
                )
            )
        )
        result = await self._execute(query)
        return result["owner"]["appliance"]["sensor"]

    async def get_location_data(self) -> Dict[str, Any]:
        """ Returns location details """
        ds = DSLSchema(self._client.schema)
        query = dsl_gql(DSLQuery(location_fragment(ds)))
        result = await self._execute(query)
        return result["owner"]["locations"]

    async def set_smart_mode(self, device_id: str, mode: SmartModeKind, is_enabled: bool) -> Dict[str, Any]:
        ds = DSLSchema(self._client.schema)
        name: DSLField = None
        if mode == SmartModeKind.Sleep:
            name = ds.Mutation.applySleepMode(applianceId=device_id,isEnabled=is_enabled,fanMode=FanMode.Lowest)
        elif mode == SmartModeKind.Turndown:
            name = ds.Mutation.applyTurndownMode(applianceId=device_id,isEnabled=is_enabled,fanMode=FanMode.Highest)
        elif mode == SmartModeKind.Whitenoise:
            name = ds.Mutation.applyWhitenoiseMode(applianceId=device_id,isEnabled=is_enabled,fanMode=FanMode.Medium)
        elif mode == SmartModeKind.Housekeeper:
            name = ds.Mutation.applyHousekeeperMode(applianceId=device_id,isEnabled=is_enabled)
        elif mode == SmartModeKind.Quiet:
            name = ds.Mutation.applyQuietMode(applianceId=device_id,isEnabled=is_enabled)
        elif mode == SmartModeKind.Quarantine:
            name = ds.Mutation.applyQuarantineMode(applianceId=device_id,isEnabled=is_enabled)
        elif mode == SmartModeKind.PowerSaver:
            name = ds.Mutation.applyPowerSaverMode(applianceId=device_id,isEnabled=is_enabled)
        elif mode == SmartModeKind.ChildLock:
            name = ds.Mutation.applyChildLockMode(applianceId=device_id,isEnabled=is_enabled)

        mutation = DSLMutation(name.select(*appliance_fields_fragment(ds)))
        result = await self._execute(dsl_gql(mutation))

        return list(result.values())[0]

    async def set_sound_mode(self, device_id: str, mode: SoundsConfig) -> Dict[str, Any]:
        ds = DSLSchema(self._client.schema)
        mutation = DSLMutation(
            ds.Mutation.applySoundsConfig(applianceId=device_id, soundsConfig=mode).select(
                *appliance_fields_fragment(ds)
            )
        )
        result = await self._execute(dsl_gql(mutation))

        return list(result.values())[0]

    async def set_automagic_mode(self, room_id: int) -> None:
        ds = DSLSchema(self._client.schema)
        mutation = DSLMutation(
            ds.Mutation.applyRoomAutomagicMode(roomId=room_id).select(
                ds.Room.id
            )
        )
        result = await self._execute(dsl_gql(mutation))

    async def set_manual_mode(self, room_id: int, fan_speed: int, target_aqi: int = 10) -> None:
        ds = DSLSchema(self._client.schema)
        mutation = DSLMutation(
            ds.Mutation.applyRoomManualMode(roomId=room_id,fanSpeed=fan_speed,targetAqi=target_aqi).select(
                ds.Room.id
            )
        )
        result = await self._execute(dsl_gql(mutation))

    async def force_room_data(self, room_id: int) -> None:
        ds = DSLSchema(self._client.schema)
        mutation = DSLMutation(
            ds.Mutation.forceRoomData(roomId=room_id).select(
                ds.Room.id
            )
        )
        result = await self._execute(dsl_gql(mutation))       


from dataclasses import dataclass
from datetime import datetime
from typing import List
from .enums import *

@dataclass
class Account:
    email: str
    firstName: str
    lastName: str
    #assume 1 location (multiple locations can't be added currently)
    address: str
    city: str
    country: str
    lat: float
    lon: float
    timezone: str
    environment_kind: EnvironmentKind
    home_kind: HomeKind
    home_size: HouseSize
    home_age: HouseAge
    home_bedrooms: HouseBedrooms

@dataclass
class Room:
    kind: RoomKind
    name: str
    size: int
    bedtime_start: str
    bedtime_end: str
    soundsConfig: SoundsConfig

@dataclass
class ApplianceFilter:
    kind: FilterKind
    days_left: int
    installed_at: datetime
    calibrated_at: datetime

@dataclass
class ApplianceState:
  firmware_version: str
  firmware_hash: str
  wifi_rssi: int
  raw_mode: int
  modes: List[ApplianceMode]

  actual_mode: ApplianceMode # none = offline

@dataclass
class SmartModes:
  quiet: bool
  housekeeper: bool
  quarantine: bool
  sleep: bool
  turndown: bool
  whitenoise: bool
  powerSaver: bool
  childLock: bool

@dataclass
class ApplianceSensor:
    kind: ApplianceSensorKind
    measured_at: datetime
    value: float

@dataclass
class Appliance:
    id: str #mac
    room: Room
    name: str
    state: ApplianceState
    filter: ApplianceFilter
    smart_modes: SmartModes
    sensors: List[ApplianceSensor]

@dataclass
class OutdoorStation:
    id: str
    name: str
    lat: float
    lon: float
    measured_at: datetime
    value: float

@dataclass
class PollenStation:
    id: str
    name: str
    measured_at: datetime
    trees: str #PollenIndex, but has None, so doesn't map right
    weeds: str #PollenIndex, but has None, so doesn't map right
    grass: str #PollenIndex, but has None, so doesn't map right
    mold: str #PollenIndex, but has None, so doesn't map right



from enum import auto
from strenum import StrEnum

class ApplianceMode(StrEnum):
  Automagic = auto()
  Sleep = auto()
  Housekeeper = auto()
  DeepClean = auto()
  Quiet = auto()
  Turndown = auto()
  WhiteNoise = auto()
  Quarantine = auto()
  Manual = auto()
  Safeguard = auto()
  PowerSaver = auto()

class SmartModeKind(StrEnum):
  Housekeeper = auto()
  Quiet = auto()
  Quarantine = auto()
  Sleep = auto()
  Turndown = auto()
  Whitenoise = auto()
  PowerSaver = auto()  
  ChildLock = auto()  

class ApplianceSensorKind(StrEnum):
  Ach = auto()
  FanSpeed = auto()
  Ttc = auto()
  Aqi = auto()
  Pm1 = auto()
  Pm2_5 = auto()
  Pm10 = auto()
  Voc = auto()
  Humidity = auto()
  Temperature = auto()
  Co2 = auto()
  Co = auto()

class EnvironmentKind(StrEnum):
  Urban = auto()
  Suburban = auto()
  Rural = auto()

class FanMode(StrEnum):
  Lowest = auto()
  Low = auto()
  Medium = auto()
  High = auto()
  Highest = auto()

class FilterKind(StrEnum):
  MamaToBe = auto()
  BasicBreather = auto()
  BigSneeze = auto()
  HomeWrecker = auto()
  CritterCuddler = auto()
  Overreactor = auto()
  RookieParent = auto()

class HomeKind(StrEnum):
  House = auto()
  Apartment = auto()

class HouseAge(StrEnum):
  Old = auto()
  New = auto()

class HouseBedrooms(StrEnum):
  One = auto()
  Two = auto()
  Three = auto()
  Four = auto()
  FiveOrMore = auto()

class HouseSize(StrEnum):
  Large = auto()
  Medium = auto()
  Small = auto()

class OutdoorStationSensorKind(StrEnum):
  Pm2_5 = auto()

class RoomKind(StrEnum):
  LivingRoom = auto()
  Kitchen = auto()
  Laundry = auto()
  HomeOffice = auto()
  Entry = auto()
  TvRoom = auto()
  MainBedroom = auto()
  KidsBedroom = auto()
  GuestBedroom = auto()
  Studio = auto()

class SoundsConfig(StrEnum):
  Enabled = auto()
  DaytimeOnly = auto()
  Disabled = auto()

class AggregateFunction(StrEnum):
  Mean = auto()
  Max = auto()
  Min = auto()
  Last = auto()



from gql import Client
from gql.utilities import update_schema_enum, update_schema_scalar

from .enums import *
from .queries import *
from .scalars import *
from .transport import AuthenticatedAIOHTTPTransport

def register_enums(client: Client):
    update_schema_enum(client.schema, "ApplianceMode", ApplianceMode)
    update_schema_enum(client.schema, "ApplianceSensorKind", ApplianceSensorKind)
    update_schema_enum(client.schema, "EnvironmentKind", EnvironmentKind)
    update_schema_enum(client.schema, "FanMode", FanMode)
    update_schema_enum(client.schema, "FilterKind", FilterKind)
    update_schema_enum(client.schema, "HomeKind", HomeKind)
    update_schema_enum(client.schema, "HouseAge", HouseAge)
    update_schema_enum(client.schema, "HouseBedrooms", HouseBedrooms)
    update_schema_enum(client.schema, "OutdoorStationSensorKind", OutdoorStationSensorKind)
    #update_schema_enum(client.schema, "PollenIndex", PollenIndex) #can't use this one because "None" is a key
    update_schema_enum(client.schema, "RoomKind", RoomKind)
    update_schema_enum(client.schema, "SoundsConfig", SoundsConfig)

def register_scalers(client: Client):
    update_schema_scalar(client.schema, "Date", DateScalar)
    update_schema_scalar(client.schema, "EpochSecond", EpochSecondScalar)



""" Mila constants """

AUTH_BASE_URL = "https://id.milacares.com"
AUTH_OIDC_TOKEN_ENDPOINT = AUTH_BASE_URL + "/auth/realms/prod/protocol/openid-connect/token"
AUTH_HEADER = "Authorization"
AUTH_CLIENT_ID = "prod-ui"
AUTH_SCOPES = "email profile"
AUTH_HEADER = "Authorization"

API_BASE_URL = "https://api.milacares.com/graphql"
API_DEFAULT_429_WAIT = 30



"""Support for Milacares authentication."""

from __future__ import annotations
import logging
from types import TracebackType
from typing import Callable, Optional, Type

from oauthlib.oauth2 import LegacyApplicationClient
from async_oauthlib import OAuth2Session
from milasdk.const import AUTH_CLIENT_ID, AUTH_OIDC_TOKEN_ENDPOINT, AUTH_SCOPES
from milasdk.exceptions import OAuthError

LOG = logging.getLogger(__name__)

class MilaOauth2:
    """
    Handle authentication with Oidc.  Mila appears to use the legacy flow for its application
    authentication, so we'll implement that instead of the usual flows that don't handle passwords
    directly.
    """

    def __init__(
        self,
        token: dict[str, str] | None = None,
        token_updater: Callable[[str], None] | None = None,
    ) -> None:
        """Initialize self.
        Keyword Arguments:
            token {Optional[Dict[str, str]]} -- Authorization token (default: {None})
            token_updater {Optional[Callable[[str], None]]} -- Callback when the token is updated (default: {None})
            scope {Optional[str]} -- List of scopes (default: "email profile")
        """
        self.client_id = AUTH_CLIENT_ID
        self.scope = AUTH_SCOPES
        self.token_updater = token_updater

        self.extra = {"client_id": self.client_id}

        self._oauth = OAuth2Session(
            client = LegacyApplicationClient(self.client_id),
            token=token,
            token_updater=self.token_updater,
            scope=self.scope
        )

    async def async_refresh_token(self) -> dict[str, str | int]:
        """Refresh and return new token."""
        try:
            token = await self._oauth.refresh_token(AUTH_OIDC_TOKEN_ENDPOINT, **self.extra)

            if self.token_updater is not None:
                self.token_updater(token)

            return token
        except Exception as ex:
            raise OAuthError(f"Refresh token failure: {ex}") from ex        

    async def async_request_token(
        self,
        username: str | None = None,
        password: str | None = None,
    ) -> dict[str, str | int]:
        """
        Generic method for fetching a Mila access token.
        :param username: Your Milacares username
        :param password: Your Milacares password
        :return: A token dict
        """
        try:
            return await self._oauth.fetch_token(
                AUTH_OIDC_TOKEN_ENDPOINT,
                username=username,
                password=password,
                include_client_id=True,
            )
        except Exception as ex:
            raise OAuthError(f"Request token failure: {ex}") from ex

    @property
    def closed(self):
        return self._oauth.closed

    async def close(self):
        if not self._oauth.closed:
            await self._oauth.close()

    async def __aenter__(self) -> "MilaOauth2":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()


from datetime import datetime
from typing import Any, Dict, Optional

from graphql import GraphQLScalarType, ValueNode
from graphql.utilities import value_from_ast_untyped

# Date: An ISO-8601 encoded date string in format "yyyy-MM-dd". For instance, "2021-04-12".

def serialize_date(value: Any) -> str:
    return value.date.isoformat()

def parse_date_value(value: Any) -> datetime:
    return datetime.fromisoformat(value)

def parse_date_literal(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> datetime:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_date_value(ast_value)

DateScalar = GraphQLScalarType(
    name="Date",
    serialize=serialize_date,
    parse_value=parse_date_value,
    parse_literal=parse_date_literal,
)

# EpochSecond: The unix time stamp. The number of seconds between a particular instant and the Unix Epoch of 1970-01-01T00:00:00Z.

def serialize_epoch(value: Any) -> str:
    return int(round(value.timestamp()))

def parse_epoch_value(value: Any) -> datetime:
    return datetime.fromtimestamp(value)

def parse_epoch_literal(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> datetime:
    ast_value = value_from_ast_untyped(value_node, variables)
    return parse_epoch_value(ast_value)

EpochSecondScalar = GraphQLScalarType(
    name="Date",
    serialize=serialize_epoch,
    parse_value=parse_epoch_value,
    parse_literal=parse_epoch_literal,
)


"""Support for MilaCares authentication, generic interface."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any
from aiohttp import ClientError, ClientResponse, ClientSession

from ..const import AUTH_HEADER
from ..exceptions import MilaError

_LOGGER = logging.getLogger(__name__)

class AbstractAsyncSession(ABC):
    """Abstract class to make authenticated async requests."""

    def __init__(self, session: ClientSession) -> None:
        """Initialize the auth."""
        self._session = session

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def post(
        self,
        url: str,
        data: Any = None,
        **kwargs
    ) -> ClientResponse:
        """Wrapper for authenticated post requests."""
        try:
            access_token = await self.async_get_access_token()
        except ClientError as err:
            raise MilaError(f"Access token failure: {err}") from err
        headers = {AUTH_HEADER: f"Bearer {access_token}"}

        resp : ClientResponse = None

        return await self._session.post(
            url,
            headers=headers,
            **kwargs
        )


"""Support for MilaCares authentication, default session."""

from operator import truediv
import time
from types import TracebackType
from typing import Optional, Type, cast
from aiohttp import ClientSession

from .mila_oath2 import MilaOauth2
from .abstract_async_session import AbstractAsyncSession

CLOCK_SKEW = 5

class DefaultAsyncSession(AbstractAsyncSession):
    """Default class to make authenticated async requests."""

    def __init__(self, session: ClientSession, username: str = None, password: str = None, auth: MilaOauth2 = None, close_session: bool = True) -> None:
        """Initialize the auth."""
        super().__init__(session)
        self._auth = auth if auth else MilaOauth2()
        self._close_session = close_session
        self._token: dict = None
        self._username = username
        self._password = password

    @property
    def token(self) -> dict:
        """Return the token."""
        return self._token

    @property
    def valid_token(self) -> bool:
        """Return if token is still valid."""
        if self.token is None:
          return False

        return (
            cast(float, self.token["expires_at"])
            > time.time()
        )

    async def async_ensure_token_valid(self) -> None:
        """Ensure that the current token is valid."""
        if self.valid_token:
            return

        if self.token is None or cast(float, self.token["refresh_expires_at"] > time.time()):
            self._token = {}
            new_token = await self._auth.async_request_token(self._username, self._password)
            new_token["expires_at"] = time.time() + new_token["expires_in"] - CLOCK_SKEW
            new_token["refresh_expires_at"] = time.time() + new_token["refresh_expires_in"] - CLOCK_SKEW
        else:
            new_token = await self._auth.async_refresh_token()       
            new_token["expires_at"] = time.time() + new_token["expires_in"] - CLOCK_SKEW

        self._token.update(new_token)

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        await self.async_ensure_token_valid()
        return self.token["access_token"]

    @property
    def closed(self):
        return self._auth.closed and (self._session.closed or not self._close_session)

    async def close(self):
        if not self._auth.closed:
            await self._auth.close()
        if not self._session.closed and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "MilaOauth2":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()


import pathlib
from setuptools import setup, find_namespace_packages

try:
    import re2 as re
except ImportError:
    import re

base_package = "milasdk"

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# Pull the version from __init__.py so we don't need to maintain it in multiple places
init_txt = (HERE / base_package / "__init__.py").read_text("utf-8")
try:
    version = re.findall(r"^__version__ = ['\"]([^'\"]+)['\"]\r?$", init_txt, re.M)[0]
except IndexError:
    raise RuntimeError("Unable to determine version.")

# This call to setup() does all the work
setup(
    name="milasdklite",
    version=version,
    description="Python SDK for Mila Air Purifiers",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/tombrien/milasdklite",
    author="Tom brien",
    author_email="tom@tombrien.io",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_namespace_packages(include=[base_package, f"{base_package}*"]),
    package_data={"": ["*.gql"]},
    include_package_data=True,
    install_requires=[
        "aiohttp",
        "magicattr",
        "gql",
        "StrEnum",
        "oauthlib",
        "async_oauthlib",
    ],
)



import datetime
from gql.dsl import *

from .enums import ApplianceSensorKind, OutdoorStationSensorKind

def appliance_fields_fragment(ds: DSLSchema):
    return (
        ds.Appliance.id,
        ds.Appliance.name,
        ds.Appliance.room.select(
            ds.Room.id,
            ds.Room.kind,
            ds.Room.name,
            ds.Room.size,
            ds.Room.soundsConfig,
            ds.Room.bedtime.select(
                ds.Bedtime.localStart,
                ds.Bedtime.localEnd
            )
        ),
        ds.Appliance.state.select(
            ds.ApplianceState.firmware.select(
                ds.Firmware.version,
                ds.Firmware.hash 
            ),
            ds.ApplianceState.wifiRssi,
            ds.ApplianceState.rawMode,
            ds.ApplianceState.modes,
            ds.ApplianceState.actualMode
        ),
        ds.Appliance.filter.select(
            ds.ApplianceFilter.kind,
            ds.ApplianceFilter.installedAt,
            ds.ApplianceFilter.calibratedAt
        ),
        ds.Appliance.sensors(kinds=[e.value for e in ApplianceSensorKind]).select(
            *appliance_sensor_fields_fragment(ds)
        )
    )

def appliance_sensor_fields_fragment(ds: DSLSchema):
    return (
        ds.ApplianceSensor.kind,
        ds.ApplianceSensor.latest(precision={ "unit": "Minute", "value": "1" }).select(
            ds.InstantValue.instant,
            ds.InstantValue.value
        )
    )

def location_fragment(ds: DSLSchema):

    #I'm taking a guess here - there appear to be gaps in the pollen data
    #so, we can't just take today's value.  I'm taking the last value
    #for a given week, hopefully it works as expected
    now = datetime.datetime.utcnow()
    window = {
        "range": {
            "start": now - datetime.timedelta(days=6),
            "stop": now
        },
        "every": {
            "value": 7,
            "unit": "Day"
        },
        "fn": "Last"
    }

    return ds.Query.owner.select(
        ds.Owner.locations.select(
            ds.Location.id,
            ds.Location.address.select(
                ds.LocationAddress.city,
                ds.LocationAddress.country,
                ds.LocationAddress.point.select(
                    ds.LatLng.lat,
                    ds.LatLng.lon
                )
            ),
            ds.Location.environmentKind,
            ds.Location.homeKind,
            ds.Location.houseSize,
            ds.Location.houseAge,
            ds.Location.houseBedrooms,                        
            ds.Location.outdoorStation.select(
                ds.OutdoorStation.id,
                ds.OutdoorStation.name,
                ds.OutdoorStation.point.select(
                    ds.LatLng.lat,
                    ds.LatLng.lon
                ),
                ds.OutdoorStation.sensor(kind=OutdoorStationSensorKind.Pm2_5).select(
                    *outdoor_sensor_fields_fragment(ds)
                )
            ),
            ds.Location.pollenStation.select(
                ds.PollenStation.name #,
                # ds.PollenStation.aggregateWindow(input=window).select(
                #     ds.DailyPollenStatus.date,
                #     ds.DailyPollenStatus.status.select(
                #         ds.PollenStatus.trees,
                #         ds.PollenStatus.weeds,
                #         ds.PollenStatus.grass,
                #         ds.PollenStatus.mold
                #     )
                # )
            ),
            ds.Location.timezone
        )
    )

def outdoor_sensor_fields_fragment(ds: DSLSchema):
    return (
        ds.OutdoorStationSensor.kind,
        ds.OutdoorStationSensor.latest(precision={ "unit": "Day", "value": "1" }).select(
            ds.InstantValue.instant,
            ds.InstantValue.value
        )
    )
